package syntax.expressions;
import parsing.*;

import java.math.BigInteger;


public class OpCall extends Expression 
{
	public Symbol op;
	public Expression lhs;
	public Expression rhs;

	public OpCall(Symbol op, Expression lhs, Expression rhs)
	{
		this.op = op;
		this.lhs = lhs;
		this.rhs = rhs;
		
		OpDef opDef = OpDef.operators.get(op.name);
		if (opDef == null)
			throw new ParsingException(op, "Unknown operator: '" + op.name + "'.");
		
		if (lhs.type != null && !opDef.lhsType.equals("?") && !opDef.lhsType.equals(lhs.type))
			throw new ParsingException(lhs, "The left hand side of '" + op.name + "' expects an '" + opDef.lhsType + "' but a '" + lhs.type + "' was found.");
		if (rhs.type != null && !opDef.lhsType.equals("?") && !opDef.rhsType.equals(rhs.type))
			throw new ParsingException(rhs, "The right hand side of '" + op.name + "' expects an '" + opDef.rhsType + "' but a '" + rhs.type + "' was found.");
		
		this.type = opDef.resType;

		this.start = lhs.start;
		this.end = rhs.end;
	}

	/*
	The types are assumed to be correct.
	*/
	public Value eval()
	{
		PrimitiveValue vTrue = new PrimitiveValue("Bool", new byte[]{1}, lhs.start, rhs.end);
		PrimitiveValue vFalse = new PrimitiveValue("Bool", new byte[]{0}, rhs.start, rhs.end);
			
		if (op.name.equals("+") | op.name.equals("-") | op.name.equals("*") | op.name.equals("/") | op.name.equals("^"))
		{
			VInt vl = (VInt) lhs.eval();
			VInt vr = (VInt) rhs.eval();
			
			BigInteger ans;
			if (op.name.equals("+"))
				ans = vl.val.add(vr.val);
			else if (op.name.equals("-"))
				ans = vl.val.add(vr.val.negate());
			else if (op.name.equals("*"))
				ans = vl.val.multiply(vr.val);
			else if (op.name.equals("/"))
				ans = vl.val.divide(vr.val);
			else if (op.name.equals("^"))
				ans = vl.val.pow(vr.val.intValue());
			else
				throw new RuntimeException("Cannot occur.");

			return new VInt(ans, vl.start, vr.end);
		}
		else if (op.name.equals("++"))
		{
			VString vl = (VString) lhs.eval();
			VString vr = (VString) rhs.eval();
			
			return new VString(vl.string + vr.string, vl.start, vr.end);
			
		}
		else if (op.name.equals("&&") | op.name.equals("||"))
		{
			PrimitiveValue vl = (PrimitiveValue) lhs.eval();
			PrimitiveValue vr = (PrimitiveValue) rhs.eval();
			
			if (op.name.equals("&&"))
			{
				if (vl.bytes[0] == 1 && vr.bytes[0] == 1)
					return vTrue;
				else
					return vFalse;
			}
			if (op.name.equals("||"))
			{
				if (vl.bytes[0] == 1 || vr.bytes[0] == 1)
					return vTrue;
				else
					return vFalse;
			}
			
			throw new RuntimeException("Cannot occur.");
		}
		else if (op.name.equals("<") | op.name.equals("<=") | op.name.equals(">") | op.name.equals(">="))
		{
			VInt vl = (VInt) lhs.eval();
			VInt vr = (VInt) rhs.eval();
			
			int comp = vl.val.compareTo(vr.val);
			
			if (op.name.equals("<"))
			{
				if (comp < 0)
					return vTrue;
				else
					return vFalse;
			}
			if (op.name.equals("<="))
			{
				if (comp <= 0)
					return vTrue;
				else
					return vFalse;
			}
			if (op.name.equals(">"))
			{
				if (comp > 0)
					return vTrue;
				else
					return vFalse;
			}
			if (op.name.equals(">="))
			{
				if (comp >= 0)
					return vTrue;
				else
					return vFalse;
			}
			
			throw new RuntimeException("Cannot occur.");
		}
		else if (op.name.equals("=="))
		{
			Value vl = lhs.eval();
			Value vr = rhs.eval();
			
			if (vl.equals(vr))
				return vTrue;
			else
				return vFalse;
		}
		else if (op.name.equals("><"))
		{
			Value vl = lhs.eval();
			Value vr = rhs.eval();
			
			if (vl.equals(vr))
				return vFalse;
			else
				return vTrue;
		}
		else
		{
			throw new RuntimeException("Operator '" + op + "' not implemented.");
		}
	}

	


	public String toString()
	{
		return "(" + lhs + " " + op + " " + rhs + ")";
	}
}
